import pandas as pd
import numpy as np
import openml
import math

from abc import abstractmethod
from pathlib import Path
from typing import Optional, SupportsIndex, Tuple, Union, ClassVar

import lucid
import lucid.nn as nn

from lucid.transforms import Compose
from lucid.data import Dataset
from lucid._tensor import Tensor


__all__ = ["MNIST", "FashionMNIST"]


class _DatasetBase(Dataset):

    OPENML_ID: ClassVar[int]

    def __init__(
        self,
        root: Union[str, Path],
        train: Optional[bool] = True,
        download: Optional[bool] = False,
        transform: Optional[nn.Module | Compose] = None,
        target_transform: Optional[nn.Module | Compose] = None,
        test_size: float = 0.2,
        to_tensor: bool = True,
    ) -> None:
        self.root = Path(root)
        self.train = train
        self.transform = transform
        self.target_transform = target_transform
        self.test_size = test_size
        self.to_tensor = to_tensor
        self.root.mkdir(parents=True, exist_ok=True)

        if download:
            self._download()

        if self.train:
            self.data, self.targets = self._load_data("train")
        else:
            self.data, self.targets = self._load_data("test")

    @abstractmethod
    def _download(self) -> None: ...

    @abstractmethod
    def _load_data(self, *args, **kwargs) -> Tuple[Tensor, ...]: ...

    def __len__(self) -> int:
        return len(self.data)


class MNIST(_DatasetBase):

    OPENML_ID: ClassVar[int] = 554

    def _download(self) -> None:
        try:
            dataset = openml.datasets.get_dataset(self.OPENML_ID)
            df, _, _, _ = dataset.get_data(dataset_format="dataframe")
            df.to_csv(self.root / "MNIST.csv", index=False)

        except Exception as e:
            raise RuntimeError(f"Failed to download the MNIST dataset. Error: {e}")

    def _load_data(self, split: str) -> Tuple[Tensor, Tensor]:
        csv_path = self.root / "MNIST.csv"
        if not csv_path.exists():
            raise RuntimeError(
                f"MNIST dataset CSV file not found at {csv_path}. "
                + "Use `download=True`."
            )

        df = pd.read_csv(csv_path)
        labels = df["class"].values.astype(np.int64)
        images = df.drop(columns=["class"]).values.astype(np.float32)
        images = images.reshape(-1, 1, 28, 28)

        train_size = int(math.floor(len(images) * (1 - self.test_size)))
        if split == "train":
            images, labels = images[:train_size], labels[:train_size]
        else:
            images, labels = images[train_size:], labels[train_size:]

        if self.to_tensor:
            images = lucid.to_tensor(images, dtype=np.float32)
            labels = lucid.to_tensor(labels, dtype=np.int64)

        return images, labels

    def __getitem__(self, index: SupportsIndex) -> Tuple[Tensor, Tensor]:
        image = self.data[index].reshape(-1, 1, 28, 28)
        label = self.targets[index]

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)

        return image, label


class FashionMNIST(_DatasetBase):

    OPENML_ID: ClassVar[int] = 40996

    def _download(self) -> None:
        try:
            dataset = openml.datasets.get_dataset(self.OPENML_ID)
            df, _, _, _ = dataset.get_data(dataset_format="dataframe")
            df.to_csv(self.root / "FashionMNIST.csv", index=False)
        except Exception as e:
            raise RuntimeError(
                f"Failed to download the FashionMNIST dataset. Error: {e}"
            )

    def _load_data(self, split: str) -> Tuple[Tensor, Tensor]:
        csv_path = self.root / "FashionMNIST.csv"
        if not csv_path.exists():
            raise RuntimeError(
                f"FashionMNIST dataset CSV file not found at {csv_path}. "
                + "Use `download=True`."
            )

        df = pd.read_csv(csv_path)
        labels = df["class"].values.astype(np.int64)
        images = df.drop(columns=["class"]).values.astype(np.float32)
        images = images.reshape(-1, 1, 28, 28)

        train_size = int(math.floor(len(images) * (1 - self.test_size)))
        if split == "train":
            images, labels = images[:train_size], labels[:train_size]
        else:
            images, labels = images[train_size:], labels[train_size:]

        if self.to_tensor:
            images = lucid.to_tensor(images, dtype=np.float32)
            labels = lucid.to_tensor(labels, dtype=np.int64)

        return images, labels

    def __getitem__(self, index: SupportsIndex) -> Tuple[Tensor, Tensor]:
        image = self.data[index].reshape(-1, 1, 28, 28)
        label = self.targets[index]

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)

        return image, label



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
import lucid.nn.functional
import lucid.nn.parameter


__all__ = [
    "Dropout",
    "Dropout1d",
    "Dropout2d",
    "Dropout3d",
    "AlphaDropout",
]


class _DropoutNd(nn.Module):
    def __init__(self, p: float = 0.5) -> None:
        super().__init__()
        if p < 0 or p > 1:
            raise ValueError(
                f"Dropout probability must be between 0 and 1, but got {p}."
            )
        self.p = p


class Dropout(_DropoutNd):
    def forward(self, input_: Tensor) -> Tensor:
        return F.dropout(input_, self.p, self.training)


class Dropout1d(_DropoutNd):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=3)
        return F.dropout1d(input_, self.p, self.training)


class Dropout2d(_DropoutNd):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=4)
        return F.dropout2d(input_, self.p, self.training)


class Dropout3d(_DropoutNd):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=5)
        return F.dropout3d(input_, self.p, self.training)


class AlphaDropout(_DropoutNd):
    def forward(self, input_: Tensor) -> Tensor:
        return F.alpha_dropout(input_, self.p, self.training)



import lucid
import lucid.nn as nn

from lucid._tensor import Tensor
from lucid.types import _ShapeLike


def summarize(
    model: nn.Module,
    input_shape: _ShapeLike,
    recurse: bool = True,
    truncate_from: int | None = None,
    test_backward: bool = False,
) -> None:
    PIPELINE: str = r"│   "
    BRANCH: str = r"├── "

    def _register_hook(module: nn.Module, depth: int) -> None:

        def _hook(_module: nn.Module, input_: Tensor, output: Tensor) -> None:
            layer_name = type(_module).__name__
            input_shape = input_[0].shape if isinstance(input_, tuple) else input_.shape
            output_shape = output.shape if isinstance(output, Tensor) else None
            param_size = _module.parameter_size
            layer_count = len(_module._modules)

            if depth == 1:
                layer_name = BRANCH + layer_name
            elif depth > 1:
                layer_name = PIPELINE * (depth - 2) + BRANCH + layer_name
            if len(layer_name) > 30:
                layer_name = layer_name[:26] + "..."

            summary_ = dict(
                layer_name=layer_name,
                input_shape=input_shape,
                output_shape=output_shape,
                param_size=param_size,
                layer_count=layer_count,
            )
            module_summary.append(summary_)

        hooks.append(module.register_forward_hook(_hook))

    def _recursive_register(module: nn.Module, depth: int = 0) -> None:
        _register_hook(module, depth)
        for _, submodule in module._modules.items():
            if recurse:
                _recursive_register(submodule, depth=depth + 1)

    hooks = []
    module_summary = []

    _recursive_register(module=model)

    dummy_input = lucid.zeros(input_shape)
    out = model(dummy_input)
    if test_backward:
        out.backward()

    module_summary.reverse()

    title = f"Summary of {type(model).__name__}"
    if model._alt_name:
        title += f"({model._alt_name})"

    print(f"{title:^95}")
    print("=" * 95)
    print(f"{"Layer":<30}{"Input Shape":<25}", end="")
    print(f"{"Output Shape":<25}{"Parameter Size":<12}")
    print("=" * 95)

    total_layers = sum(layer["layer_count"] for layer in module_summary)
    total_params = model.parameter_size

    if truncate_from is not None:
        truncated_lines = len(module_summary) - truncate_from
        if truncated_lines > 0:
            module_summary = module_summary[:truncate_from]

    for layer in module_summary:
        print(
            f"{layer['layer_name']:<30}{str(layer['input_shape']):<25}",
            f"{str(layer['output_shape']):<25}",
            sep="",
            end="",
        )
        if layer["param_size"]:
            print(f"{layer['param_size']:<12,}")
        else:
            print("-")

    if truncate_from is not None and truncated_lines > 0:
        print(f"\n{f"... and more {truncated_lines} layer(s)":^95}")

    print("=" * 95)
    print(f"Total Layers(Submodules): {total_layers:,}")
    print(f"Total Parameters: {total_params:,}")
    print("=" * 95)

    for hook in hooks:
        hook()



from .lenet import *
from .alex import *
from .zfnet import *
from .vgg import *
from .inception import *
from .inception_res import *
from .resnet import *
from .resnext import *
from .senet import *
from .sknet import *



from typing import overload

import lucid
from lucid.random import _func
from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _Scalar


def seed(seed: int) -> None:
    return _func.seed(seed)


@overload
def rand(
    *shape: int, requires_grad: bool = False, keep_grad: bool = False
) -> Tensor: ...


@overload
def rand(
    shape: _ShapeLike, requires_grad: bool = False, keep_grad: bool = False
) -> Tensor: ...


def rand(*args: int, requires_grad: bool = False, keep_grad: bool = False) -> Tensor:
    shape = lucid._get_overloaded_shape(args)
    return _func.rand(shape, requires_grad=requires_grad, keep_grad=keep_grad)


def randint(
    low: int,
    high: int | None,
    size: int | _ShapeLike = 1,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return _func.randint(low, high, size, requires_grad, keep_grad)


@overload
def randn(
    *shape: int, requires_grad: bool = False, keep_grad: bool = False
) -> Tensor: ...


@overload
def randn(
    shape: _ShapeLike, requires_grad: bool = False, keep_grad: bool = False
) -> Tensor: ...


def randn(
    *args: int | _ShapeLike, requires_grad: bool = False, keep_grad: bool = False
) -> Tensor:
    shape = lucid._get_overloaded_shape(args)
    return _func.randn(shape, requires_grad=requires_grad, keep_grad=keep_grad)


def uniform(
    low: _Scalar = 0,
    high: _Scalar = 1,
    size: int | _ShapeLike = 1,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return _func.uniform(low, high, size, requires_grad, keep_grad)
